#include "dxut.h"
#include "HeightMap.h"

float Noise(int x)
{
	x = (x<<13) ^ x;
	return (1.0f - ((x * (x*x * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.f);    
}

float CosInterpolate(float v1, float v2, float a)
{
	float angle = a * D3DX_PI;
	float prc = (1.0f - cos(angle)) * 0.5f;
	return  v1*(1.0f - prc) + v2*prc;
}

HeightMap::HeightMap( Point size, float maxheight )
{
	LOG_DEBUG( "HeightMap constructer called. Size: " << size.x << "," << size.y << " Max height: " << maxheight );
	try
	{
		mSize = size;
		mMaxHeight = maxheight;

		mHeightTable = new float[mSize.x * mSize.y];
		memset(mHeightTable, 0, sizeof(float)*mSize.x*mSize.y);
	}
	catch(...)
	{
		LOG_ERROR( "Error initializing heightmap" );
	}
}

HeightMap::~HeightMap()
{
	Release();
}

void HeightMap::operator*=(const HeightMap &rhs)
{
	for(int y = 0; y < mSize.y; y++)
	{
		for(int x = 0; x < mSize.x; x++)
		{
			float a = mHeightTable[x + y * mSize.x] / mMaxHeight;
			float b = 1.0f;
			if(x <= rhs.mSize.x && y <= rhs.mSize.y)
				b = rhs.mHeightTable[x + y * mSize.x] / rhs.mMaxHeight;

			mHeightTable[x + y * mSize.x] = a * b * mMaxHeight;
		}
	}

	mNewMinHeight = mMaxHeight;
	mNewMaxHeight = 0.f;

	// Compute new min/max height
	for(int y = 0; y < mSize.y; y++)
	{
		for(int x = 0; x < mSize.x; x++)
		{
			float height = mHeightTable[x + y * mSize.x];
			mNewMinHeight = min( height, mNewMinHeight );
			mNewMaxHeight = max( height, mNewMaxHeight );
		}
	}
	LOG_DEBUG( "Height map combined. New height min: " << mNewMinHeight << " max: " << mNewMaxHeight );
}

void HeightMap::Release()
{
	SAFE_DELETE_ARRAY( mHeightTable );	
}

HRESULT HeightMap::LoadFromFile(LPDIRECT3DDEVICE9 device, char fileName[])
{
	try
	{
		//Reset the HeightMap to 0.0f
		memset(mHeightTable, 0, sizeof(float) * mSize.x * mSize.y);

		//Initiate the texture variables
		LPDIRECT3DTEXTURE9 heightMapTexture = NULL;
		D3DXIMAGE_INFO info;

		// Load the texture (and scale it to our height map mSize)
		if(FAILED(D3DXCreateTextureFromFileExA(device, fileName, mSize.x, mSize.y, 1, D3DUSAGE_DYNAMIC, 
			D3DFMT_L8, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 
			NULL, &info, NULL, &heightMapTexture)))return E_FAIL;

		// Lock the texture
		D3DLOCKED_RECT sRect;
		heightMapTexture->LockRect(0, &sRect, NULL, NULL);
		BYTE *bytes = (BYTE*)sRect.pBits;

		// Extract height values from the texture
		for(int y = 0; y < mSize.y; y++)
		{
			for(int x = 0; x < mSize.x; x++)
			{
				BYTE *b = bytes + y * sRect.Pitch + x;
				float height = (*b / 255.0f)  * mMaxHeight;
				mHeightTable[x + y * mSize.x] = height;

			}
		}

			// Unlock the texture
			heightMapTexture->UnlockRect(0);

			// Release texture
			heightMapTexture->Release();
	}
	catch(...)
	{
		LOG_ERROR( "Error loading height map data" );
	}

	LOG_INFO( "New height map created. Height min: " << mNewMinHeight << " max: " << mNewMaxHeight );

	return S_OK;
}

HRESULT HeightMap::CreateRandomHeightMap(int seed, float noiseSize, float persistence, int octaves)
{
	//For each map node
	for(int y=0;y<mSize.y;y++)
		for(int x=0;x<mSize.x;x++)
		{
			//Scale x & y to the range of 0.0 - mSize
			float xf = ((float)x / (float)mSize.x) * noiseSize;
			float yf = ((float)y / (float)mSize.y) * noiseSize;

			float total = 0;			

			// For each octave
			for(int i =0; i < octaves; i++)
			{
				//Calculate frequency and amplitude (different for each octave)
				float freq = pow(2.0f, i);
				float amp = pow(persistence, i);

				//Calculate the x,y noise coordinates
				float tx = xf * freq;
				float ty = yf * freq;
				int tx_int = (int)tx;
				int ty_int = (int)ty;

				//Calculate the fractions of x & y
				float fracX = tx - tx_int;
				float fracY = ty - ty_int;

				//Get the noise of this octave for each of these 4 points
				float v1 = Noise(tx_int + ty_int * 57 + seed);
				float v2 = Noise(tx_int+ 1 + ty_int * 57 + seed);
				float v3 = Noise(tx_int + (ty_int+1) * 57 + seed);
				float v4 = Noise(tx_int + 1 + (ty_int+1) * 57 + seed);

				//Smooth in the X-axis
				float i1 = CosInterpolate(v1 , v2 , fracX);
				float i2 = CosInterpolate(v3 , v4 , fracX);

				//Smooth in the Y-axis
				total += CosInterpolate(i1 , i2 , fracY) * amp;
			}

			int b = (int)(128.f + total * 128.f);
			if(b < 0)b = 0;
			if(b > 255)b = 255;

			// Save to height map table
			float height = (b / 255.0f) * mMaxHeight;			
			mHeightTable[x + y * mSize.x] = height;								
		}

		mNewMinHeight = mMaxHeight;
		mNewMaxHeight = 0.f;

		// Compute new min/max height
		for(int y=0;y<mSize.y;y++)
		{
			for(int x=0;x<mSize.x;x++)
			{
				float height = mHeightTable[x + y * mSize.x];
				mNewMinHeight = min( height, mNewMinHeight );
				mNewMaxHeight = max( height, mNewMaxHeight );
			}
		}
		LOG_INFO( "New height map created. Height min: " << mNewMinHeight << " max: " << mNewMaxHeight );

		return S_OK;
}

HRESULT HeightMap::GenerateHeightMapTexture( LPDIRECT3DDEVICE9 device, LPDIRECT3DTEXTURE9 *texture )
{
	if(FAILED(D3DXCreateTexture(device, mSize.x, mSize.y, 1, D3DUSAGE_DYNAMIC, 
		D3DFMT_L8, D3DPOOL_DEFAULT, texture)))
	{
		LOG_ERROR( "Failed to create height map texture" );
		return E_FAIL;
	};

	//Lock the texture
	D3DLOCKED_RECT sRect;
	((LPDIRECT3DTEXTURE9)*texture)->LockRect(0, &sRect, NULL, NULL);
	BYTE *bytes = (BYTE*)sRect.pBits;

	//Extract height values from the texture
	for(int y = 0; y < mSize.y; y++)
	{
		for(int x = 0; x < mSize.x; x++)
		{
			BYTE *bDest = (BYTE*)sRect.pBits;
			bDest += y * sRect.Pitch + x;

			// Clamp to 1 - 255 since texture can only have between these values
			int b = (int)((mHeightTable[x + y * mSize.x] / mMaxHeight) * 255.f);

			if(b < 0) b = 0;
			if(b > 255) b = 255;

			*bDest = b;
		}
	}

	//Unlock the texture
	((LPDIRECT3DTEXTURE9)*texture)->UnlockRect(0);

	return S_OK;
}

void HeightMap::Raise(RECT r, float f)
{
	for(int y = r.top; y <= r.bottom; y++)
	{
		for(int x = r.left; x <= r.right; x++)
		{
			mHeightTable[x + y * mSize.x] += f;
			if(mHeightTable[x + y * mSize.x] < -mMaxHeight)mHeightTable[x + y * mSize.x] = -mMaxHeight;
			if(mHeightTable[x + y * mSize.x] > mMaxHeight)mHeightTable[x + y * mSize.x] = mMaxHeight;
		}
	}
}

void HeightMap::Raise( Point pt, int radius, float f )
{
	float RadiusPowerTwo = pow((float)radius, 2);	

	for(int x = -radius; x < radius; x++)
	{
		for(int y = -radius; y < radius; y++)
		{
			if ( x*x + y*y <= RadiusPowerTwo)
			{
				//you are in the circle, what distance are you from the centre?
				// float distance = sqrt(pow((float)x,2) + pow((float)y,2));	
				int indexx = pt.x + x;
				int indexy = pt.y + y;

				mHeightTable[ indexx + indexy * mSize.x] += f;	
				
				//if(mHeightTable[ indexx + indexy * mSize.x] < mNewMinHeight)mHeightTable[indexx + indexy * mSize.x] = mNewMinHeight;
				//if(mHeightTable[ indexx + indexy * mSize.x] > mNewMaxHeight)mHeightTable[indexx + indexy * mSize.x] = mNewMaxHeight;
			}
		}
	}
}

void HeightMap::Smooth()
{
	//Create temporary HeightMap
	float *hm = new float[mSize.x * mSize.y];
	memset(hm, 0, sizeof(float) * mSize.x * mSize.y);

	for(int y = 0; y < mSize.y; y++) 
	{
		for(int x = 0; x < mSize.x; x++)
		{
			float totalHeight = 0.0f;
			int noNodes = 0;

			for(int y1 = y-1; y1 <= y+1; y1++)
			{
				for(int x1 = x-1; x1 <= x+1; x1++)
				{
					if(x1 >= 0 && x1 < mSize.x && y1 >= 0 && y1 < mSize.y)
					{
						totalHeight += mHeightTable[x1 + y1 * mSize.x];
						noNodes++;
					}

					hm[x + y * mSize.x] = totalHeight / (float)noNodes;
				}
			}
		}
	}

	//Replace old HeightMap with smoothed HeightMap
	SAFE_DELETE_ARRAY( mHeightTable );	
	mHeightTable = hm;
}

void HeightMap::Cap(float capHeight)
{
	//Cap terrain to capHeight
	mMaxHeight = 0.0f;

	for(int y = 0; y < mSize.y; y++) 
	{
		for(int x = 0; x < mSize.x; x++)
		{
			mHeightTable[x + y * mSize.x] -= capHeight;
			if(mHeightTable[x + y * mSize.x] < 0.0f)
				mHeightTable[x + y * mSize.x] = 0.0f;

			if(mHeightTable[x + y * mSize.x] > mMaxHeight)
				mMaxHeight = mHeightTable[x + y * mSize.x];
		}
	}
}

bool HeightMap::Within(Point p)
{
		return p.x >= 0 && p.y >= 0 && p.x < mSize.x && p.y < mSize.y;
}